/**
 * Copyright 2008, Daniel Molina Cabrera <danimolina@gmail.com>
 * 
 * This file is part of software Realea
 * 
 * Realea is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Realea is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ma2lsia.h"
#include "selectlsimp.h"
#include <cassert>
#include <cmath>
#include <cstdio>

using namespace realea;
using namespace realea::internal;

void MA2LSIA::setEffortRatio(double ratio) {
	if (ratio == 1)
	   throw new string("MA2LSIA::effortRatio is not valide");

	m_effort = ratio;
}

MA2LSIA::MA2LSIA(IEAlgorithm *alg, ILocalSearch *ls) : Hybrid(alg,ls), m_memory(NULL) {
   m_effort = -1;
   m_restart = NULL;
   m_debug = false;
   m_select_improvement = NULL;
   m_frec_update = 5;
}

void MA2LSIA::setDebug(void) {
    m_debug = true;
}

MA2LSIA::~MA2LSIA(void) {
    if (m_memory) {
	delete m_memory;
    }

    if (m_restart) {
	delete m_restart;
    }

    if (m_select_improvement) {
	delete m_select_improvement;
    }

    if (m_initeval) {
	delete m_initeval;
    }
}

unsigned MA2LSIA::calculateFrec(double nevalalg, double nevalls, unsigned intensity, unsigned evalprev, double ratio) {

    assert(ratio > 0);

    double coc = ratio*(nevalalg + nevalls + intensity) - evalprev;
    double div = (1-ratio);

    if (div == 0) {
	throw string("MA2LSIA::Ratio is too low");
    }

    return ((unsigned) floor(coc/div));
}

unsigned MA2LSIA::init(void) {
    initLs();

    m_initMaxEval = m_running->maxEval();
    unsigned neval = m_alg->init();

    if (m_select_improvement == NULL) {
	m_select_improvement = new SelectBestToImprove();
    }

    if (m_memory == NULL) {
	m_memory = new LSParametersMemory(m_alg->getPop()->size());
	m_alg->getPop()->setObserver(m_memory);
    }

    m_nevalalg = neval;
    m_nevalls = 0;
    return neval;
}

void MA2LSIA::setMaxEval(unsigned int maxeval) {
    assert(maxeval >= m_intensity);
    unsigned frec = calculateFrec(m_nevalalg, m_nevalls, m_intensity, m_nevalalg, m_effort);
    m_alg->setMaxEval(frec);
}

void MA2LSIA::setRestart(RestartStrategy *restart, int maxfailed) {

    if (m_restart) {
	delete m_restart;

    }

    m_restart = restart;
    m_maxfailed = maxfailed;
}

void MA2LSIA::setRunning(Running *running) {
    ProxyEA::setRunning(running);
    unsigned frec = calculateFrec(m_nevalalg, m_nevalls, m_intensity, m_nevalalg, m_effort);
    m_alg->setRunning(m_running->getSubRunning(frec));
}

void MA2LSIA::updateFrec(double ratio_effort, double ratio_improvement) {
    // Improvement of LS is lesser than expected
    if (ratio_effort != ratio_improvement) {
        double C = ratio_effort/ratio_improvement;
	unsigned frec = (unsigned) ceil(C*m_intensity/(1-C));

	if (frec < 50) {
	    frec = 50;
	}
	else if (frec > 3000) {
	    frec = 3000;
	}

	printf("New MaxEval: %u\n", frec);
	m_alg->setMaxEval(frec);
    }
}



void MA2LSIA::recoverIndividual(unsigned oldind, tGen *aind, unsigned size, tGen *aoptional, unsigned size_optional) {
    m_alg->recoverIndividual(oldind, aind, size-1, aoptional, size_optional);

    if (aind[size]) {
        m_alg->getPop()->getInd(oldind)->incremCount("non_improved");
    }

    if (aoptional != NULL) {
        IParallelLocalSearch *ls = (IParallelLocalSearch *) m_ls;
	ILSParameters *params = ls->recoverOptions(aoptional, size_optional);
	assert(m_memory);
	m_memory->store(oldind, params);
    }

}

void MA2LSIA::storeIndividual(tIndividualRealPtr ind, tGen **paind, unsigned *pmax, tGen **paoptional, unsigned *psize_optional) {
    tGen *asol, *aoptional_sol;
    tGen *asol_ma;
    unsigned size_sol, size_optional;
    tGen *aparams;
    unsigned size_param;
    unsigned size;

    m_alg->storeIndividual(ind, &asol, &size_sol, &aoptional_sol, &size_optional);
    assert(size_optional == 0 && aoptional_sol == NULL);

    asol_ma = new tGen[size_sol+1];
    copy(asol, asol+size_sol, asol_ma);
    delete[] asol;
    asol_ma[size_sol] = (ind->getCount("non_improved") > 0) ? 1 : 0;

    *paind = asol_ma;
    *pmax = size_sol+1;

    size_param = 0;
    
    if (m_memory) {
        unsigned posind = ind->getId();
        IParallelLocalSearch *ls = (IParallelLocalSearch *) m_ls;
	ILSParameters *params = m_memory->recover(posind);
	ls->storeOptions(params, &aparams, &size_param);
    }

    size = size_optional+size_param;
    assert(size > 0);

    *psize_optional = size_param;
    *paoptional = NULL;

    if (aoptional_sol != NULL || aparams != NULL) {
       *paoptional = new tGen[size];

       if (aoptional_sol != NULL) {
	  copy(aoptional_sol, aoptional_sol+size_optional, *paoptional);
	  delete[] aoptional_sol;
       }

       if (aparams != NULL) {
	  copy(aparams, aparams+size_param, *paoptional+size_optional);
	  delete[] aparams;
       }

    }
}

void MA2LSIA::setDif(bool debug, string ident, unsigned id, tFitness oldfit, tFitness newfit) {
   if (debug) {

      if (oldfit!= newfit) {
	 printf("%s[%2d]:\t%e-> %e\t\t%e\n", ident.c_str(), id, oldfit, newfit, fabs(newfit-oldfit));
      }
      else {
	 printf("%s[%2d]:\t%e\n", ident.c_str(), id, newfit);
      }
   }

}

bool MA2LSIA::hasDiversity(PopulationReal *pop) {
    return true;
    double percen[5];
    pop->getPercentils(percen, 4);


	    printf("EA::Improvement: %e\t%e\t%e\t%e\t%e\n", percen[0], percen[1], 
			percen[2], percen[3], percen[4]);


    if (percen[2] == percen[4]) {
	return false;
    }
    else if (percen[1] == percen[3]) {
	return false;
    }
    else if (fabs((percen[0] - percen[2])/percen[2]) < 1e-3) {
	return false;
    }
    else {
	return true;    
    }
}

void MA2LSIA::setInitEval(IEval*eval) {
   Hybrid::setInitEval(eval);
   m_initeval = (IEvalInd *) new EvalRunning(eval, m_running);
}

unsigned MA2LSIA::realApply(tChromosomeReal &bestsol, tFitness &bestfitness) {
    tIndividualReal *ind, *best;
    unsigned posind;
    tFitness oldfitness, fitness, fitness_alg;
    unsigned ndim = bestsol.size();
    tChromosomeReal sol(ndim), sol_alg(ndim);
    unsigned alg_failed;
    
    PopulationReal *pop_alg = m_alg->getPop();
    deque<tIndividualReal*> ind_to_improve;
    tFitness improvement_alg, improvement_ls;
    tFitness improvement_alg_step, improvement_ls_step;
    double nevalalg_step, nevalls_step;
    tFitness improvement_ratio, old_improvement_ratio, improvement;
    unsigned inc_neval;

    sol=bestsol;
    fitness = bestfitness;
    improvement_alg = improvement_ls = 0;
    improvement_alg_step = improvement_ls_step = 0;
    nevalalg_step = nevalls_step = 0;

    unsigned initMax = m_running->numEval();
    fitness_alg = pop_alg->getInd(pop_alg->getBest())->perf();

    improvement_ratio = m_effort;

    alg_failed = 0;
    unsigned n_itera = 0;

//    while ( (m_nevalalg+m_nevalls) < m_initMaxEval && !m_running->hasFoundOptime()) {
    while (!m_running->isFinish()) {
        tFitness old_fitness = fitness_alg;
	inc_neval = m_alg->realApply(sol_alg, fitness_alg);
	m_nevalalg += inc_neval; nevalalg_step += inc_neval;
	improvement = fabs(fitness_alg-old_fitness);
	improvement_alg += improvement; improvement_alg_step += improvement;

	if (fitness_alg == old_fitness) {
	    alg_failed++;
	}
	else {
	    alg_failed = 0;
	}

	if (m_debug) {
	    setDif(m_debug, "Alg", pop_alg->getBest(), old_fitness, fitness_alg);
	}

	// Check the optime
	if (m_running->isOptime(fitness_alg)) {
	    continue;
	}

	// Select the new individual to increm
	m_select_improvement->getIndsToImprove(pop_alg, ind_to_improve);

	if (ind_to_improve.size()!=0) {
	    // Select the individual to improve
	    posind = m_select_improvement->selectIndToImprove(ind_to_improve);
	}
	else {
	    // Choose the new one randomly
	    posind = m_random->randint(0, pop_alg->size()-1);
	}

	ind = pop_alg->getInd(posind);
	sol = ind->sol();

	ILSParameters *params = m_memory->recover(posind);
	bool recover = (params != NULL);
	
	if (params == NULL) {	
	    // Apply the LS to the best one with the rest of intensity
	    params = m_ls->getInitOptions(sol);
	}
	
	fitness = ind->perf();
	oldfitness = fitness;

	inc_neval = m_ls->apply(params, sol, fitness, m_intensity);

	m_nevalls += inc_neval;
	nevalls_step += inc_neval;

	ind->incremCount("ls"); 
	improvement = fabs(fitness-oldfitness);
	improvement_ls += improvement;
	improvement_ls_step += improvement;

	setDif(m_debug, "LS ", ind->getId(), oldfitness, fitness);

	if (m_problem->isBetter(fitness, fitness_alg)) {
	    fitness_alg = fitness;
	}

	if (!m_problem->isBetter(fitness, oldfitness)) {
	   ind->incremCount("non_improved");

	   if (recover) {
	      m_memory->remove(posind);
	   }
	   else {
	      delete params;
	   }

	   if (ind_to_improve.empty() && m_restart != NULL && !m_running->isFinish()) {
		m_restart->apply(pop_alg, m_problem, m_initeval);
		
		if (m_debug) {
		    printf("Restart_NotImprovement\t: %e\n", fitness);
		}
	   }
	}
	else {
	    pop_alg->change(posind, sol, fitness);
	    m_memory->store(posind, params);
	}

	if (m_restart != NULL && m_maxfailed > 0) {
	     if (alg_failed > m_maxfailed && !hasDiversity(pop_alg) && !m_running->isFinish()) {
		m_restart->apply(pop_alg, m_problem, m_initeval);
		
		if (m_debug) {
		    printf("Restart_AlgFailed\t: %e\n", fitness_alg);
		}
		alg_failed = 0;
	     }
	}

	n_itera ++;

	if (n_itera == m_frec_update) {
	    old_improvement_ratio = improvement_ratio;
	    improvement_ratio = improvement_alg_step/(improvement_alg_step+improvement_ls_step);
	    updateFrec(old_improvement_ratio, improvement_ratio);
	    nevalalg_step = nevalls_step = 0;
	    improvement_alg_step = improvement_ls_step = 0;
	    n_itera = 0;
	}
    }

    if (m_debug) {
        double ratio_effort = ( (double)m_nevalalg)/(m_nevalalg+m_nevalls);
	printf("RatioEffort Alg/LS: [%.0f/%.0f]\n", 100*ratio_effort, 100*(1-ratio_effort));

        double ratio_alg = improvement_alg/(improvement_alg+improvement_ls);
	printf("RatioImprovement Alg/LS: [%.0f/%.0f]\n", 100*ratio_alg, 100*(1-ratio_alg));
    }
    unsigned neval = m_running->numEval()-initMax;
    best = pop_alg->getInd(pop_alg->getBest());
    bestsol = best->sol();
    bestfitness = best->perf();
    m_running->reset();
    m_alg->setMaxEval(m_intensity);
    return neval;
}
